import * as React from 'react';
import { createContext, createElement } from 'react';
import poseFactory from 'popmotion-pose';
import {
  DomPopmotionPoser,
  DomPopmotionConfig
} from 'popmotion-pose/lib/types';
import {
  ChildRegistration,
  CurrentPose,
  PoseContextProps,
  PoseElementInternalProps,
  PopStyle
} from './PoseElement.types';
import { warning, invariant } from 'hey-listen';

export const PoseParentContext = createContext({});

type Ref = (ref: Element) => any;
type RefSetters = {
  ref?: Ref;
  innerRef?: Ref;
  hostRef?: Ref;
};

const calcPopFromFlowStyle = (el: HTMLElement): PopStyle => {
  const { offsetTop, offsetLeft, offsetWidth, offsetHeight } = el;

  return {
    position: 'absolute',
    top: offsetTop,
    left: offsetLeft,
    width: offsetWidth,
    height: offsetHeight
  };
};

const hasPose = (pose: CurrentPose, key: string) =>
  Array.isArray(pose) ? pose.indexOf(key) !== -1 : pose === key;

const objectToMap = (obj: { [key: string]: any }): Map<string, any> =>
  Object.keys(obj).reduce((map, key) => {
    map.set(key, { raw: obj[key] });
    return map;
  }, new Map());

class PoseElement extends React.PureComponent<PoseElementInternalProps> {
  props: PoseElementInternalProps;
  poser: DomPopmotionPoser;
  ref: Element;
  styleProps: { [key: string]: any };
  children: Set<ChildRegistration> = new Set();
  popStyle?: PopStyle;

  /**
   * Children handlers
   * =============================================
   */
  private childrenHandlers: PoseContextProps = {
    registerChild: props => {
      this.children.add(props);
      if (this.poser) this.flushChildren();
    },
    onUnmount: child => this.poser.removeChild(child),
    getParentPoseConfig: () => this.props.poseConfig,
    getInitialPoseFromParent: () => this.getInitialPose()
  };

  /**
   * Lifecycle
   * =============================================
   */
  getInitialPose(): CurrentPose {
    const { getInitialPoseFromParent, pose, _pose, initialPose } = this.props;

    if (initialPose) {
      return initialPose;
    } else {
      // Feels like this could probably be simpler
      const parentPose = getInitialPoseFromParent && getInitialPoseFromParent();
      const thisPose = Array.isArray(pose) ? pose : [pose];
      const thisInternalPose = Array.isArray(_pose) ? _pose : [_pose];

      return Array.isArray(parentPose)
        ? [...parentPose, ...thisPose, ...thisInternalPose]
        : [parentPose, ...thisPose, ...thisInternalPose];
    }
  }

  getFirstPose(): CurrentPose | void {
    const { initialPose, pose, _pose } = this.props;
    // If don't have an initial pose explicitly defined, return early
    if (!initialPose) return;

    const thisPose = Array.isArray(pose) ? pose : [pose];
    const thisInternalPose = Array.isArray(_pose) ? _pose : [_pose];

    return [...thisPose, ...thisInternalPose];
  }

  getSetProps() {
    const {
      children,
      elementType,
      poseConfig,
      onChange, // Deprecated for 2.0.0
      onValueChange,
      innerRef,
      _pose,
      pose,
      initialPose,
      onPoseComplete,
      getParentPoseConfig,
      registerChild,
      onUnmount,
      getInitialPoseFromParent,
      popFromFlow,
      values,
      parentValues,
      onDragStart,
      onDragEnd,
      ...props
    } = this.props;

    // If we're popping this element out from the DOM flow, build
    // and apply position: absolute styles that visually match the previous
    // location in the DOM
    if (popFromFlow && this.ref && this.ref instanceof HTMLElement) {
      if (!this.popStyle) {
        props.style = {
          ...props.style,
          ...calcPopFromFlowStyle(this.ref as HTMLElement)
        };
        this.popStyle = props.style;
      } else {
        props.style = this.popStyle;
      }
    } else {
      this.popStyle = null;
    }

    // Deprecated for 2.0.0
    // If this is a function, it's intended for the DOM element
    if (typeof onChange === 'function') props.onChange = onChange;

    return props;
  }

  /**
   * We need to get a ref to the underlying DOM element. Styled Components and
   * other libraries use `innerRef`, though this will be swallowed if the
   * styled component is not a primitive (ie styled(Component)).
   *
   * Instead we pass another function, `hostRef`, as recommended by Facebook
   * https://github.com/facebook/react/issues/11401
   *
   * We also only pass `ref` to DOM primitive components.
   */
  getRefs = (): RefSetters => {
    const refs: RefSetters = {};
    const { elementType } = this.props;

    if (typeof elementType === 'string') {
      refs.ref = this.setRef;
    } else {
      refs.innerRef = this.setRef;
      refs.hostRef = this.setRef;
    }

    return refs;
  };

  setRef = (ref: Element) => {
    if (ref instanceof Element || (this.ref && ref === null)) {
      const { innerRef } = this.props;
      if (innerRef) innerRef(ref);
      this.ref = ref;
    }
  };

  componentDidMount() {
    invariant(
      typeof this.ref !== 'undefined',
      `No DOM ref found. If you're converting an existing component via posed(Component), you must ensure you're passing the hostRef prop to your underlying DOM element.`
    );

    const {
      poseConfig,
      onChange, // Deprecated 2.0.0
      onValueChange,
      registerChild,
      values,
      parentValues,
      onDragStart,
      onDragEnd
    } = this.props;

    const config: DomPopmotionConfig = {
      ...poseConfig,
      initialPose: this.getInitialPose(),
      values: values || poseConfig.values,
      parentValues: parentValues ? objectToMap(parentValues) : undefined,
      props: this.getSetProps(),
      onDragStart,
      onDragEnd,
      onChange: onValueChange
        ? onValueChange
        : typeof onChange !== 'function' ? onChange : undefined // 2.0.0 set to just `onValueChange`
    };

    warning(
      onChange === undefined || typeof onChange === 'function',
      'The onChange prop is deprecated. Use onValueChange instead.'
    );

    // If first in tree
    if (!registerChild) {
      this.initPoser(poseFactory(this.ref, config));
    } else {
      registerChild({
        element: this.ref,
        poseConfig: config,
        onRegistered: poser => this.initPoser(poser)
      });
    }
  }

  UNSAFE_componentWillUpdate({ pose, _pose }: PoseElementInternalProps) {
    if (hasPose(pose, 'flip') || hasPose(_pose, 'flip')) this.poser.measure();
  }

  componentDidUpdate(prevProps: PoseElementInternalProps) {
    const { pose, _pose } = this.props;
    this.poser.setProps(this.getSetProps());

    if (pose !== prevProps.pose || pose === 'flip') this.setPose(pose);
    if (_pose !== prevProps._pose || _pose === 'flip') this.setPose(_pose);
  }

  componentWillUnmount() {
    if (!this.poser) return;
    const { onUnmount } = this.props;
    if (onUnmount) onUnmount(this.poser);
    this.poser.destroy();
  }

  initPoser(poser: DomPopmotionPoser) {
    this.poser = poser;
    this.flushChildren();

    const firstPose = this.getFirstPose();
    if (firstPose) this.setPose(firstPose);
  }

  setPose(pose: CurrentPose) {
    const { onPoseComplete } = this.props;
    const poseList: string[] = Array.isArray(pose) ? pose : [pose];

    Promise.all(poseList.map(key => key && this.poser.set(key))).then(
      () => onPoseComplete && onPoseComplete()
    );
  }

  flushChildren() {
    this.children.forEach(({ element, poseConfig, onRegistered }) =>
      onRegistered(this.poser.addChild(element, poseConfig))
    );

    this.children.clear();
  }

  render() {
    const { elementType, children } = this.props;

    return (
      <PoseParentContext.Provider value={this.childrenHandlers}>
        {createElement(
          elementType,
          { ...this.getSetProps(), ...this.getRefs() },
          children
        )}
      </PoseParentContext.Provider>
    );
  }
}

export { PoseElement };
